"""
Console Scripts
---------------

To see available commands
::

    wads --help


"""

import setuptools  # to avoid a warning due to distutils_patch.py
import os
import json

root_dir = os.path.dirname(__file__)
root_dir_name = os.path.basename(root_dir)
rjoin = lambda *paths: os.path.join(root_dir, *paths)

data_dir = rjoin('data')
licenses_json_path = rjoin(data_dir, 'github_licenses.json')

pkg_dir = os.path.dirname(root_dir)
pkg_join = lambda *paths: os.path.join(pkg_dir, *paths)

# TODO: Change to use ini format? (Or yaml or toml?)
wads_configs_file = rjoin(data_dir, 'wads_configs.json')
try:
    wads_configs = json.load(open(wads_configs_file))
except FileNotFoundError:
    wads_configs = {
        'populate_dflts': {
            'description': "There's a bit of an air of mystery around this project...",
            'root_url': None,
            'author': None,
            'license': 'mit',
            'description_file': 'README.md',
            'long_description': 'file:README.md',
            'long_description_content_type': 'text/markdown',
            'keywords': None,
            'install_requires': None,
            'verbose': None,
        }
    }

pkg_path_names = ('.gitignore', 'setup.py')
pkg_paths = {pkg_join(name) for name in pkg_path_names}

from wads.populate import populate_pkg_dir


def main():
    import argh  # pip install argh

    from wads.pack import argh_kwargs as pack_kw

    # from wads.docs_gen import argh_kwargs as docs_gen_kw

    parser = argh.ArghParser()
    parser.add_commands(**pack_kw)
    # parser.add_commands(**docs_gen_kw)
    parser.dispatch()


if __name__ == '__main__':
    main()



"""Utils to package and publish.



The typical sequence of the methodic and paranoid could be something like this:

::

    python pack.py current-configs  # see what you got
    python pack.py increment-configs-version  # update (increment the version and write that in setup.cfg
    python pack.py current-configs-version  # see that it worked
    python pack.py current-configs  # ... if you really want to see the whole configs again (you're really paranoid)
    python pack.py run-setup  # see that it worked
    python pack.py twine-upload-dist  # publish
    # and then go check things work...



If you're crazy (or know what you're doing) just do

::

    python pack.py go

"""
import subprocess
from setuptools import find_packages
import json
import re
import sys
from pprint import pprint
from typing import Union, Mapping, Iterable, Generator
from configparser import ConfigParser
from functools import partial
import os
import argh
import pytest
import pylint.lint

CONFIG_FILE_NAME = 'setup.cfg'
METADATA_SECTION = 'metadata'
OPTIONS_SECTION = 'options'
DFLT_OPTIONS = {
    'packages': 'find:',
    'include_package_data': True,
    'zip_safe': False,
}

pjoin = lambda *p: os.path.join(*p)
DOCSRC = 'docsrc'
DFLT_PUBLISH_DOCS_TO = None  # 'github'


def git(command='status', work_tree='.', git_dir=None):
    """Launch git commands.

    :param command: git command (e.g. 'status', 'branch', 'commit -m "blah"', 'push', etc.)
    :param work_tree: The work_tree directory (i.e. where the project is)
    :param git_dir: The .git directory (usually, and by default, will be taken to be "{work_tree}/.git/"
    :return: What ever the command line returns (decoded to string)
    """

    """

    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ add myFile
    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ commit -m 'something'

    """
    work_tree = os.path.abspath(os.path.expanduser(work_tree))
    if git_dir is None:
        git_dir = os.path.join(work_tree, '.git')
    command = f'git --git-dir="{git_dir}" --work-tree="{work_tree}" {command}'
    r = subprocess.check_output(command, shell=True)
    if isinstance(r, bytes):
        r = r.decode()
    return r.strip()


def get_name_from_configs(pkg_dir, assert_exists=True):
    """Get name from local setup.cfg (metadata section)"""
    configs = read_configs(pkg_dir=pkg_dir)
    name = configs.get('name', None)
    if assert_exists:
        assert name is not None, 'No name was found in configs'
    return name


def clog(condition, *args, log_func=pprint, **kwargs):
    if condition:
        log_func(*args, **kwargs)


Path = str


def check_in(
    commit_message: str,
    work_tree='.',
    git_dir=None,
    auto_choose_default_action=False,
    bypass_docstring_validation=False,
    bypass_tests=False,
    bypass_code_formatting=False,
    verbose=False,
):
    """Validate, normalize, stage, commit and push your local changes to a remote repository.

    :param commit_message: Your commit message
    :type commit_message: str
    :param work_tree: The relative or absolute path of the working directory. Defaults to '.'.
    :type work_tree: str, optional
    :param git_dir: The relative or absolute path of the git directory. If None, it will be taken to be "{work_tree}/.git/". Defaults to None.
    :type git_dir: str, optional
    :param auto_choose_default_action: Set to True if you don't want to be prompted and automatically select the default action. Defaults to False.
    :type auto_choose_default_action: bool, optional
    :param bypass_docstring_validation: Set to True if you don't want to check if a docstring exists for every module, class and function. Defaults to False.
    :type bypass_docstring_validation: bool, optional
    :param bypass_tests: Set to True if you don't want to run doctests and other tests. Defaults to False.
    :type bypass_tests: bool, optional
    :param bypass_code_formatting: Set to True if you don't want the code to be automatically formatted using axblack. Defaults to False.
    :type bypass_code_formatting: bool, optional
    :param verbose: Set to True if you want to log extra information during the process. Defaults to False.
    :type verbose: bool, optional
    """

    def ggit(command):
        r = git(command, work_tree=work_tree, git_dir=git_dir)
        clog(verbose, r, log_func=print)
        return r

    def print_step_title(step_title):
        if verbose:
            print()
            print(f'============= {step_title.upper()} =============')
        else:
            print(f'-- {step_title}')

    def abort():
        print('Aborting')
        sys.exit()

    def confirm(action, default):
        if auto_choose_default_action:
            return default
        return argh.interaction.confirm(action, default)

    def verify_current_changes():
        print_step_title('Verify current changes')
        if 'nothing to commit, working tree clean' in ggit('status'):
            print('No changes to check in.')
            abort()

    def pull_remote_changes():
        print_step_title('Pull remote changes')
        ggit('stash')
        result = ggit('pull')
        ggit('stash apply')
        if result != 'Already up to date.' and not confirm(
            'Your local repository was not up to date, but it is now. Do you want to continue',
            default=True,
        ):
            abort()

    def validate_docstrings():
        def run_pylint(current_dir):
            if os.path.exists(os.path.join(current_dir, '__init__.py')):
                result = pylint.lint.Run(
                    [
                        current_dir,
                        '--disable=all',
                        '--enable=C0114,C0115,C0116',
                    ],
                    do_exit=False,
                )
                if result.linter.stats['global_note'] < 10 and not confirm(
                    f'Docstrings are missing in directory {current_dir}. Do you want to continue',
                    default=True,
                ):
                    abort()
            else:
                for subdir in next(os.walk(current_dir))[1]:
                    run_pylint(os.path.join(current_dir, subdir))

        run_pylint('.')

    def run_tests():
        print_step_title('Run tests')
        args = ['--doctest-modules']
        if verbose:
            args.append('-v')
        else:
            args.extend(['-q', '--no-summary'])
        result = pytest.main(args)
        if result == pytest.ExitCode.TESTS_FAILED and not confirm(
            'Tests have failed. Do you want to push anyway', default=False
        ):
            abort()
        elif result not in [
            pytest.ExitCode.OK,
            pytest.ExitCode.NO_TESTS_COLLECTED,
        ]:
            print(
                'Something went wrong when running tests. Please check the output.'
            )
            abort()

    def format_code():
        print_step_title('Format code')
        subprocess.run(
            'black --line-length=79 .',
            shell=True,
            check=True,
            capture_output=not verbose,
        )

    def stage_changes():
        print_step_title('Stage changes')
        result = ggit('status')
        if 'Changes not staged for commit' in result:
            if not auto_choose_default_action and not verbose:
                print(result)
            if confirm(
                'Do you want to stage all your pending changes', default=True
            ):
                git('add -A')

    def commit_changes():
        print_step_title('Commit changes')
        if 'no changes added to commit' in ggit('status'):
            print('No changes to check in.')
            abort()
        ggit(f'commit --message="{commit_message}"')

    def push_changes():
        print_step_title('Push changes')
        if not confirm(
            'Your changes have been commited. Do you want to push',
            default=True,
        ):
            abort()
        ggit('push')

    try:
        verify_current_changes()
        pull_remote_changes()
        if not bypass_docstring_validation:
            validate_docstrings()
        if not bypass_tests:
            run_tests()
        if not bypass_code_formatting:
            format_code()
        stage_changes()
        commit_changes()
        push_changes()

    except subprocess.CalledProcessError:
        print('An error occured. Please check the output.')
        abort()

    print('Your changes have been checked in successfully!')


def goo(
    pkg_dir,
    commit_message,
    git_dir=None,
    auto_choose_default_action=False,
    bypass_docstring_validation=False,
    bypass_tests=False,
    bypass_code_formatting=False,
    verbose=False,
):
    """Validate, normalize, stage, commit and push your local changes to a remote repository.

    :param pkg_dir: The relative or absolute path of the working directory. Defaults to '.'.
    :type pkg_dir: str, optional
    :param commit_message: Your commit message
    :type commit_message: str
    :param git_dir: The relative or absolute path of the git directory. If None, it will be taken to be "{work_tree}/.git/". Defaults to None.
    :type git_dir: str, optional
    :param auto_choose_default_action: Set to True if you don't want to be prompted and automatically select the default action. Defaults to False.
    :type auto_choose_default_action: bool, optional
    :param bypass_docstring_validation: Set to True if you don't want to check if a docstring exists for every module, class and function. Defaults to False.
    :type bypass_docstring_validation: bool, optional
    :param bypass_tests: Set to True if you don't want to run doctests and other tests. Defaults to False.
    :type bypass_tests: bool, optional
    :param bypass_code_formatting: Set to True if you don't want the code to be automatically formatted using axblack. Defaults to False.
    :type bypass_code_formatting: bool, optional
    :param verbose: Set to True if you want to log extra information during the process. Defaults to False.
    :type verbose: bool, optional
    """

    return check_in(
        commit_message=commit_message,
        work_tree=pkg_dir,
        git_dir=git_dir,
        auto_choose_default_action=auto_choose_default_action,
        bypass_docstring_validation=bypass_docstring_validation,
        bypass_tests=bypass_tests,
        bypass_code_formatting=bypass_code_formatting,
        verbose=verbose,
    )


# TODO: Add a function that adds/commits/pushes the updated setup.cfg
# TODO: Include tests as first step
# TODO: blackify
# TODO: Git pull and make sure no conflicts before moving on...
def go(
    pkg_dir,
    version=None,
    publish_docs_to=DFLT_PUBLISH_DOCS_TO,
    verbose: bool = True,
    skip_git_commit: bool = False,
    answer_yes_to_all_prompts: bool = False,
    twine_upload_options_str: str = '',
    keep_dist_pkgs=False,
):
    """Update version, package and deploy:
    Runs in a sequence: increment_configs_version, update_setup_cfg, run_setup, twine_upload_dist

    :param version: The desired version (if not given, will increment the current version
    :param verbose: Whether to print stuff or not
    :param skip_git_commit: Whether to skip the git commit and push step
    :param answer_yes_to_all_prompts: If you do git commit and push, whether to ask confirmation after showing status

    """

    version = increment_configs_version(pkg_dir, version)
    update_setup_cfg(pkg_dir, verbose=verbose)
    run_setup(pkg_dir)
    twine_upload_dist(pkg_dir, twine_upload_options_str)

    if not keep_dist_pkgs:
        delete_pkg_directories(pkg_dir, verbose)

    if publish_docs_to:
        generate_and_publish_docs(pkg_dir, publish_docs_to)
    if not skip_git_commit:
        git_commit_and_push(
            pkg_dir, version, verbose, answer_yes_to_all_prompts
        )


def git_commit_and_push(
    pkg_dir,
    version=None,
    verbose: bool = True,
    answer_yes_to_all_prompts: bool = False,
):
    def ggit(command):
        r = git(command, work_tree=pkg_dir)
        clog(verbose, r, log_func=print)

    ggit('status')

    if not answer_yes_to_all_prompts:
        answer = input("Should I do a 'git add *'? ([Y]/n): ")
        if answer and answer != 'Y':
            print("Okay, I'll stop here.")
            return
    ggit('add *')

    ggit('status')  # show status again

    if not answer_yes_to_all_prompts:
        answer = input(f'Should I commit -m "{version}" and push? ([Y]/n)')
        if answer and answer != 'Y':
            print("Okay, I'll stop here.")
            return
    ggit(f'commit -m {version}')
    ggit(f'push')


def generate_and_publish_docs(pkg_dir, publish_docs_to='github'):
    # TODO: Figure out epythet and wads relationship -- right now, there's a reflexive dependency
    from epythet.docs_gen.autogen import make_autodocs
    from epythet.docs_gen.call_make import make

    make_autodocs(pkg_dir)
    if publish_docs_to:
        make(pkg_dir, publish_docs_to)


def delete_pkg_directories(pkg_dir: Path, verbose=True):
    from shutil import rmtree

    pkg_dir, pkg_dirname = validate_pkg_dir(pkg_dir)
    names_to_delete = ['dist', 'build', f'{pkg_dirname}.egg-info']
    for name_to_delete in names_to_delete:
        delete_dirpath = os.path.join(pkg_dir, name_to_delete)
        if os.path.isdir(delete_dirpath):
            if verbose:
                print(f'Deleting folder: {delete_dirpath}')
            rmtree(delete_dirpath)


# def update_version(version):
#     """Updates version (writes to setup.cfg)"""
#     pass
def _get_pkg_dir(pkg_dir: Path, validate=True) -> Path:
    pkg_dir, pkg_dirname = validate_pkg_dir(pkg_dir)
    if validate:
        validate_pkg_dir(pkg_dir)
    return pkg_dir


def _get_pkg_dir_and_name(pkg_dir):
    pkg_dir = os.path.realpath(pkg_dir)
    if pkg_dir.endswith(os.sep):
        pkg_dir = pkg_dir[:-1]
    pkg_dirname = os.path.basename(pkg_dir)
    return pkg_dir, pkg_dirname


def validate_pkg_dir(pkg_dir):
    """Asserts that the pkg_dir is actually one (has a pkg_name/__init__.py file)"""
    pkg_dir, pkg_dirname = _get_pkg_dir_and_name(pkg_dir)
    assert os.path.isdir(pkg_dir), f"Directory {pkg_dir} wasn't found"
    assert pkg_dirname in os.listdir(
        pkg_dir
    ), f"pkg_dir={pkg_dir} doesn't itself contain a dir named {pkg_dirname}"
    assert '__init__.py' in os.listdir(os.path.join(pkg_dir, pkg_dirname)), (
        f'pkg_dir={pkg_dir} contains a dir named {pkg_dirname}, '
        f"but that dir isn't a package (does not have a __init__.py"
    )

    return pkg_dir, pkg_dirname


def current_configs(pkg_dir):
    configs = read_configs(pkg_dir=_get_pkg_dir(pkg_dir))
    pprint(configs)


def current_configs_version(pkg_dir):
    pkg_dir = _get_pkg_dir(pkg_dir)
    return read_configs(pkg_dir=pkg_dir)['version']


# TODO: Both setup and twine are python. Change to use python objects directly.
def update_setup_cfg(pkg_dir, new_deploy=False, version=None, verbose=True):
    """Update setup.cfg (at this point, just updates the version).
    If version is not given, will ask pypi (via http request) what the current version is, and increment that.
    """
    pkg_dir = _get_pkg_dir(pkg_dir)
    configs = read_and_resolve_setup_configs(
        pkg_dir=_get_pkg_dir(pkg_dir), new_deploy=new_deploy, version=version,
    )
    pprint('\n{configs}\n')
    clog(verbose, pprint(configs))
    write_configs(pkg_dir=pkg_dir, configs=configs)


def set_version(pkg_dir, version):
    """Update setup.cfg (at this point, just updates the version).
    If version is not given, will ask pypi (via http request) what the current version is, and increment that.
    """
    pkg_dir = _get_pkg_dir(pkg_dir)
    configs = read_configs(pkg_dir)
    assert isinstance(version, str), 'version should be a string'
    configs['version'] = version
    write_configs(pkg_dir=pkg_dir, configs=configs)


def increment_configs_version(
    pkg_dir, version=None,
):
    """Update setup.cfg (at this point, just updates the version).
    If version is not given, will ask pypi (via http request) what the current version is, and increment that.
    """
    pkg_dir = _get_pkg_dir(pkg_dir)
    configs = read_configs(pkg_dir=pkg_dir)
    version = _get_version(
        pkg_dir, version=version, configs=configs, new_deploy=False
    )
    version = increment_version(version)
    configs['version'] = version
    write_configs(pkg_dir=pkg_dir, configs=configs)
    return version


def run_setup(pkg_dir):
    """Run ``python setup.py sdist bdist_wheel``"""
    print(
        '--------------------------- setup_output ---------------------------'
    )
    pkg_dir = _get_pkg_dir(pkg_dir)
    original_dir = os.getcwd()
    os.chdir(pkg_dir)
    setup_output = subprocess.run(
        f'{sys.executable} setup.py sdist bdist_wheel'.split(' ')
    )
    os.chdir(original_dir)
    # print(f"{setup_output}\n")


def twine_upload_dist(pkg_dir, options_str=None):
    """Publish to pypi. Runs ``python -m twine upload dist/*``"""
    print(
        '--------------------------- upload_output ---------------------------'
    )
    pkg_dir = _get_pkg_dir(pkg_dir)
    original_dir = os.getcwd()
    os.chdir(pkg_dir)
    # TODO: dist/*? How to publish just last one?
    if options_str:
        command = f'{sys.executable} -m twine upload {options_str} dist/*'
    else:
        command = f'{sys.executable} -m twine upload dist/*'
    # print(command)
    subprocess.run(command.split(' '))
    os.chdir(original_dir)
    # print(f"{upload_output.decode()}\n")


# TODO: A lot of work done here to read setup.cfg. setup function apparently does it for you. How to use that?

# TODO: postprocess_ini_section_items and preprocess_ini_section_items: Add comma separated possibility?
# TODO: Find out if configparse has an option to do this processing alreadys
def postprocess_ini_section_items(
    items: Union[Mapping, Iterable]
) -> Generator:
    r"""Transform newline-separated string values into actual list of strings (assuming that intent)

    >>> section_from_ini = {
    ...     'name': 'wads',
    ...     'keywords': '\n\tpackaging\n\tpublishing'
    ... }
    >>> section_for_python = dict(postprocess_ini_section_items(section_from_ini))
    >>> section_for_python
    {'name': 'wads', 'keywords': ['packaging', 'publishing']}

    """
    splitter_re = re.compile('[\n\r\t]+')
    if isinstance(items, Mapping):
        items = items.items()
    for k, v in items:
        if v.startswith('\n'):
            v = splitter_re.split(v[1:])
            v = [vv.strip() for vv in v if vv.strip()]
            v = [
                vv for vv in v if not vv.startswith('#')
            ]  # remove commented lines
        yield k, v


# TODO: Find out if configparse has an option to do this processing already
def preprocess_ini_section_items(items: Union[Mapping, Iterable]) -> Generator:
    """Transform list values into newline-separated strings, in view of writing the value to a ini formatted section

    >>> section = {
    ...     'name': 'wads',
    ...     'keywords': ['documentation', 'packaging', 'publishing']
    ... }
    >>> for_ini = dict(preprocess_ini_section_items(section))
    >>> print('keywords =' + for_ini['keywords'])  # doctest: +NORMALIZE_WHITESPACE
    keywords =
        documentation
        packaging
        publishing

    """
    if isinstance(items, Mapping):
        items = items.items()
    for k, v in items:
        if isinstance(v, str) and not v.startswith('"') and ',' in v:
            v = list(map(str.strip, v.split(',')))
        if isinstance(v, list):
            v = '\n\t' + '\n\t'.join(v)

        yield k, v


def read_configs(
    pkg_dir: Path,
    postproc=postprocess_ini_section_items,
    section=METADATA_SECTION,
):
    assert isinstance(
        pkg_dir, Path
    ), "It doesn't look like pkg_dir is a path. Did you perhaps invert pkg_dir and postproc order"
    pkg_dir = _get_pkg_dir(pkg_dir)
    config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
    c = ConfigParser()
    if os.path.isfile(config_filepath):
        c.read_file(open(config_filepath, 'r'))
        print(type(section), section)
        try:
            d = c[section]
        except KeyError:
            d = {}
        if postproc:
            d = dict(postproc(d))
    else:
        d = {}
    return d


def write_configs(
    pkg_dir: Path,
    configs,
    preproc=preprocess_ini_section_items,
    dflt_options=DFLT_OPTIONS,
):
    assert isinstance(
        pkg_dir, Path
    ), "It doesn't look like pkg_dir is a path. Did you perhaps invert pkg_dir and configs order"
    pkg_dir = _get_pkg_dir(pkg_dir)
    config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
    c = ConfigParser()
    if os.path.isfile(config_filepath):
        c.read_file(open(config_filepath, 'r'))

    metadata_dict = dict(preproc(configs))
    options = dict(
        dflt_options, **read_configs(pkg_dir, preproc, OPTIONS_SECTION)
    )

    # TODO: Legacy. Reorg key to [section][key] mapping to avoid such ugly complexities.
    for k in [
        'install_requires',
        'install-requires',
        'packages',
        'zip_safe',
        'include_package_data',
    ]:
        if k in metadata_dict:
            if k not in options:
                options[k] = metadata_dict.pop(
                    k
                )  # get it out of metadata_dict and into options
            else:
                metadata_dict.pop(
                    k
                )  # if it's both in metadata and in options, just get it out of metadata

    c[METADATA_SECTION] = metadata_dict
    c[OPTIONS_SECTION] = options
    with open(config_filepath, 'w') as fp:
        c.write(fp)


# dflt_formatter = Formatter()


def increment_version(version_str):
    version_nums = list(map(int, version_str.split('.')))
    version_nums[-1] += 1
    return '.'.join(map(str, version_nums))


try:
    import requests

    requests_is_installed = True
except ModuleNotFoundError:
    requests_is_installed = False


def http_get_json(
    url, use_requests=requests_is_installed
) -> Union[dict, None]:
    """Make ah http request to url and get json, and return as python dict"""

    if use_requests:
        import requests

        r = requests.get(url)
        if r.status_code == 200:
            return r.json()
        else:
            raise ValueError(f'response code was {r.status_code}')
    else:
        import urllib.request
        from urllib.error import HTTPError

        req = urllib.request.Request(url)
        try:
            r = urllib.request.urlopen(req)
            if r.code == 200:
                return json.loads(r.read())
            else:
                raise ValueError(f'response code was {r.code}')
        except HTTPError:
            return None  # to indicate (hopefully) that name doesn't exist
        except Exception:
            raise


DLFT_PYPI_PACKAGE_JSON_URL_TEMPLATE = (
    'https://pypi.python.org/pypi/{package}/json'
)


# TODO: Perhaps there's a safer way to analyze errors (and determine if the package exists or other HTTPError)
def current_pypi_version(
    pkg_dir: Path,
    name: Union[None, str] = None,
    url_template=DLFT_PYPI_PACKAGE_JSON_URL_TEMPLATE,
    use_requests=requests_is_installed,
) -> Union[str, None]:
    """
    Return version of package on pypi.python.org using json.

    ::

        current_pypi_version('py2store')
        '0.0.7'


    :param package: Name of the package
    :return: A version (string) or None if there was an exception (usually means there
    """
    pkg_dir, pkg_dirname = validate_pkg_dir(pkg_dir)
    name = name or get_name_from_configs(pkg_dir)
    assert (
        pkg_dirname == name
    ), f'pkg_dirname ({pkg_dirname}) and name ({name}) were not the same'
    url = url_template.format(package=name)
    t = http_get_json(url, use_requests=use_requests)
    releases = t.get('releases', [])
    if releases:
        return sorted(releases, key=lambda r: tuple(map(int, r.split('.'))))[
            -1
        ]


def next_version_for_package(
    pkg_dir: Path,
    name: Union[None, str] = None,
    url_template=DLFT_PYPI_PACKAGE_JSON_URL_TEMPLATE,
    version_if_current_version_none='0.0.1',
    use_requests=requests_is_installed,
) -> str:
    name = name or get_name_from_configs(pkg_dir=pkg_dir)
    current_version = current_pypi_version(
        name, url_template, use_requests=use_requests
    )
    if current_version is not None:
        return increment_version(current_version)
    else:
        return version_if_current_version_none


def _get_version(
    pkg_dir: Path,
    version,
    configs,
    name: Union[None, str] = None,
    new_deploy=False,
):
    version = version or configs.get('version', None)
    if version is None:
        try:
            if new_deploy:
                version = next_version_for_package(
                    pkg_dir, name
                )  # when you want to make a new package
            else:
                version = current_pypi_version(
                    pkg_dir, name
                )  # when you want to make a new package
        except Exception as e:
            print(
                f'Got an error trying to get the new version of {name} so will try to get the version from setup.cfg...'
            )
            print(f'{e}')
            version = configs.get('version', None)
            if version is None:
                raise ValueError(
                    f"Couldn't fetch the next version from PyPi (no API token?), "
                    f'nor did I find a version in setup.cfg (metadata section).'
                )
    return version


def read_and_resolve_setup_configs(
    pkg_dir: Path, new_deploy=False, version=None, assert_names=True
):
    """make setup params and call setup

    :param pkg_dir: Directory where the pkg is (which is also where the setup.cfg is)
    :param new_deploy: whether this setup for a new deployment (publishing to pypi) or not
    :param version: The version number to set this up as.
                    If not given will look at setup.cfg[metadata] for one,
                    and if not found there will use the current version (requesting pypi.org)
                    and bump it if the new_deploy flag is on
    """
    # read the config file (get a dict with it's contents)
    pkg_dir, pkg_dirname = _get_pkg_dir_and_name(pkg_dir)
    if assert_names:
        validate_pkg_dir(pkg_dir)

    configs = read_configs(pkg_dir)

    # parse out name and root_url
    assert (
        'root_url' in configs or 'url' in configs
    ), "configs didn't have a root_url or url"

    name = configs['name'] or pkg_dirname
    if assert_names:
        assert (
            name == pkg_dirname
        ), f'config name ({name}) and pkg_dirname ({pkg_dirname}) are not equal!'

    if 'root_url' in configs:
        root_url = configs['root_url']
        if root_url.endswith(
            '/'
        ):  # yes, it's a url so it's always forward slash, not the systems' slash os.sep
            root_url = root_url[:-1]
        url = f'{root_url}/{name}'
    elif 'url' in configs:
        url = configs['url']
    else:
        raise ValueError(
            f"configs didn't have a root_url or url. It should have at least one of these!"
        )

    # Note: if version is not in config, version will be None,
    #  resulting in bumping the version or making it be 0.0.1 if the package is not found (i.e. first deploy)

    meta_data_dict = {k: v for k, v in configs.items()}

    # make the setup_kwargs
    setup_kwargs = dict(
        meta_data_dict,
        # You can add more key=val pairs here if they're missing in config file
    )

    version = _get_version(pkg_dir, version, configs, name, new_deploy)

    def text_of_readme_md_file():
        try:
            with open('README.md') as f:
                return f.read()
        except:
            return ''

    dflt_kwargs = dict(
        name=f'{name}',
        version=f'{version}',
        url=url,
        packages=find_packages(),
        include_package_data=True,
        platforms='any',
        # long_description=text_of_readme_md_file(),
        # long_description_content_type="text/markdown",
        description_file='README.md',
    )

    configs = dict(dflt_kwargs, **setup_kwargs)

    return configs


argh_kwargs = {
    'namespace': 'pack',
    'functions': [
        generate_and_publish_docs,
        current_configs,
        increment_configs_version,
        current_configs_version,
        twine_upload_dist,
        read_and_resolve_setup_configs,
        update_setup_cfg,
        go,
        goo,
        check_in,
        get_name_from_configs,
        run_setup,
        current_pypi_version,
        validate_pkg_dir,
        git_commit_and_push,
    ],
    'namespace_kwargs': {
        'title': 'Package Configurations',
        'description': 'Utils to package and publish.',
    },
}


def main():
    import argh  # pip install argh

    argh.dispatch_commands(argh_kwargs.get('functions', None))


if __name__ == '__main__':
    main()



import re
from itertools import chain
import pkgutil
from functools import partial
import os
import subprocess

standard_lib_dir = os.path.dirname(os.__file__)


def git(command='status', work_tree='.', git_dir=None):
    """Launch git commands.

    :param command: git command (e.g. 'status', 'branch', 'commit -m "blah"', 'push', etc.)
    :param work_tree: The work_tree directory (i.e. where the project is)
    :param git_dir: The .git directory (usually, and by default, will be taken to be "{work_tree}/.git/"
    :return: What ever the command line returns (decoded to string)
    """

    """
    
    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ add myFile
    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ commit -m 'something'

    """
    work_tree = os.path.abspath(os.path.expanduser(work_tree))
    if git_dir is None:
        git_dir = os.path.join(work_tree, '.git')
    command = f'git --git-dir="{git_dir}" --work-tree="{work_tree}" {command}'
    return subprocess.check_output(command, shell=True).decode()


def is_standard_lib_path(path):
    return path.startswith(standard_lib_dir)


def standard_lib_module_names(
    is_standard_lib_path=is_standard_lib_path,
    name_filt=lambda name: not name.startswith('_'),
):
    return filter(
        name_filt,
        (
            module_info.name
            for module_info in pkgutil.iter_modules()
            if is_standard_lib_path(module_info.module_finder.path)
        ),
    )


def mk_conditional_logger(condition, func=print):
    if condition:
        return func
    else:

        def do_nothing(*args, **kwargs):
            pass

        return do_nothing


def clog(condition, func=print, *args, **kwargs):
    if condition:
        func(*args, **kwargs)


def mk_replacer_from_dict(from_to_dict):
    """Make a function that does multiple replacements (in a single pass).

    >>> r = mk_replacer_from_dict({'is': 'are', 'life': 'butterflies'})
    >>> r("There is no life in the void.")
    'There are no butterflies in the void.'

    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :return: A replacer function that you can apply to strings to carry out the replacements
    """

    p = re.compile(
        '|'.join(map(r'({})'.format, map(re.escape, from_to_dict.keys())))
    )
    f = lambda x: from_to_dict[x.group(0)]

    def replacer(s):
        return p.sub(f, s)

    return replacer


def mk_import_root_replacer(from_to_dict):
    """Make a function that does multiple import name replacements.

    For a use case, see replace_import_names. This is just a helper function.

    >>> replace = mk_import_root_replacer({'foo': 'FOO', 'bar': 'BAR'})
    >>>
    >>> assert replace('from foo import BLAH') == 'from FOO import BLAH'
    >>> assert replace('from foo.bar import BLAH') == 'from FOO.bar import BLAH'
    >>> assert replace('import bar') == 'import BAR'

    Partial matches are not replaced (that's a good thing!):

    >>> assert replace('import barmitzvah as oy') == 'import barmitzvah as oy'
    >>> assert replace('from foobar import hello') == 'from foobar import hello'

    Yes, and it works with dotpaths:

    >>> replace = mk_import_root_replacer({'where.it.was': 'where.it.is.now'})
    >>> replace('import where.it.was as here')
    'import where.it.is.now as here'

    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :return: A replacer function that you can apply to strings to carry out the replacements

    """
    t = dict(
        chain.from_iterable(
            [
                (f'(?<=from\ ){old}(?=[\.\ ])', f'{new}'),
                (f'(?<=import\ ){old}(?=[\.\s])', f'{new}'),
            ]
            for old, new in from_to_dict.items()
        )
    )

    p = re.compile('|'.join(map(r'({})'.format, t.keys())))
    f = lambda x: from_to_dict[x.group(0)]

    def replacer(s):
        ss = p.sub(f, s + ' ')
        return ss[:-1]

    return replacer


def replace_import_names(
    source_store,
    from_to_dict,
    target_store=None,
    dryrun=True,
    verbose=True,
    replacer_factory=mk_import_root_replacer,
    add_comment_at_the_end_of_lines_replaced=False,
):
    """Replace import names.

    Use case: You've renamed something or moved some modules (remember UNIX? Same as move!) and have to go through
    all your files and notebooks and replace those names.
    Now, if you have a nice IDE, we suggest you use refactoring instead -- as long as you have any uses in the scope.
    But sometimes it's not enough. You might have text/html documments, or jupyter notebooks, etc.
    So you can use this instead.

    Be warned though:
    - You should look at the pattern that is used to match, and make sure it won't create havoc.
    - Backup your documents so you can revert!
    - Print matches before you actually apply them all in bulk.
    - Use at your own risk!

    For examples, see the mk_import_root_replacer helper function.

    :param source_store:
    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :param target_store: The store
    :param dryrun:
    :param verbose:
    :param replacer_factory: makes the replacer = replacer_factory(from_to_dict)
    :param add_comment_at_the_end_of_lines_replaced: True/False or an actual string to add at the end of replaced lines
    :return:
    """

    suffix = ''
    if add_comment_at_the_end_of_lines_replaced is False:
        suffix = ''
    elif add_comment_at_the_end_of_lines_replaced is True:
        suffix = '# line_was_edited_by_wads'
    elif isinstance(add_comment_at_the_end_of_lines_replaced, str):
        suffix = add_comment_at_the_end_of_lines_replaced
        if not suffix.startswith('#'):
            suffix = '# ' + suffix
    else:
        raise TypeError(
            f"Don't know what to do with such a type of add_comment_at_the_end_of_lines_replaced: "
            f'{add_comment_at_the_end_of_lines_replaced}'
        )

    _clog = mk_conditional_logger(condition=verbose, func=print)

    if target_store is None:
        target_store = {}  # use a dict to write results
        _clog(
            "You didn't specify a target_store, so I'll write all of this in a dict and return it to you!"
        )

    if source_store == target_store:
        # TODO: Add confirmation (user input) to protect more. Make this an option (so total automatic is possible)
        _clog(
            'I just wanted you to be aware that you are using the same store for source and target.'
            "This means I'm about to overwrite your files!!"
        )
        if dryrun:
            _clog("... Right now, dryrun=True, so I'm just pretending!")

    replacer = replacer_factory(from_to_dict)

    if dryrun:
        replace_prompt = 'will replace'
    else:
        replace_prompt = '   replacing'
    with____prompt = '        with'

    for k, v in source_store.items():

        lines_replaced = 0

        def gen():
            nonlocal lines_replaced
            for i, line in enumerate(
                v.split('\n'), 1
            ):  # TODO: Double traversal. Find online splitter
                new_line = replacer(line) + suffix
                yield new_line
                if (
                    new_line != line
                ):  # TODO: Double traversal... could have replacer return a "replaced_something" flag
                    lines_replaced += 1
                    clog(
                        verbose or dryrun,
                        print,
                        f'{k}:{i}:\n{replace_prompt}\t{line}\n{with____prompt}\t{new_line}',
                    )

        new_v = '\n'.join(gen())

        if not dryrun and lines_replaced > 0:
            target_store[k] = new_v

    return target_store


fc = dict(
    reset='\033[0m',  # alias for reset_all
    reset_all='\033[0m',
    bold='\033[1m',
    dim='\033[2m',
    underlined='\033[4m',
    blink='\033[5m',
    reverse='\033[7m',
    hidden='\033[8m',
    reset_bold='\033[21m',
    reset_dim='\033[22m',
    reset_underlined='\033[24m',
    reset_blink='\033[25m',
    reset_reverse='\033[27m',
    reset_hidden='\033[28m',
    default='\033[39m',
    black='\033[30m',
    red='\033[31m',
    green='\033[32m',
    yellow='\033[33m',
    blue='\033[34m',
    magenta='\033[35m',
    cyan='\033[36m',
    gray='\033[37m',
    dark_gray='\033[90m',
    dark_red='\033[91m',
    dark_green='\033[92m',
    dark_yellow='\033[93m',
    dark_blue='\033[94m',
    dark_magenta='\033[95m',
    dark_cyan='\033[96m',
    white='\033[97m',
    background_default='\033[49m',
    background_black='\033[40m',
    background_red='\033[41m',
    background_green='\033[42m',
    background_yellow='\033[43m',
    background_blue='\033[44m',
    background_magenta='\033[45m',
    background_cyan='\033[46m',
    background_gray='\033[47m',
    background_dark_gray='\033[100m',
    background_dark_red='\033[101m',
    background_dark_green='\033[102m',
    background_dark_yellow='\033[103m',
    background_dark_blue='\033[104m',
    background_dark_magenta='\033[105m',
    background_dark_cyan='\033[106m',
    background_white='\033[107m',
)

try:
    from py2store.sources import DictAttr

    fc = DictAttr(**fc)
except ModuleNotFoundError:
    pass  # DictAttr is convenient, but not necessary


def highlight(
    string,
    effect=fc['reverse'],
    beg_mark='[[',
    end_mark=']]',
    end_effect=fc['reset_all'],
):
    r"""Interprets a string's highlight markers to be able to make highlights in the string.

    This is meant for very simple situations. A more powerful and fast function could be made by
    using regular expressions and a map to map "codes" to "effects".

    Try this:

    >>> print(highlight("This is [[the section]] that is [[highlighted]]."))  # doctest:+SKIP

    Above, "reverse" is used as the default effect.
    But You can change that to bold blue ink on yellow background. That's three effects:
    1 (for bold), 34, for the blue foreground (ink), and 43 for the "yellow" (more like brown)
    background (paper).

    >>> my_string = "This is [[the section]] that is [[highlighted]]."
    >>> print(highlight(my_string, "\033[1;34;43m"))  # doctest:+SKIP

    \033[whaaaa?!? Yeah... well, either you do it that has-no-life-outside-unicode way.
    If so, Ansi help you!
    See [this wiki section](https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_parameters).
    or [this tutorial](https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#rich-text).

    If not, we've prepared a map between human language and effect codes in the form of the
    `fc` variable of this module. It's a dict (and if you have ``py2store``, it's a mapping containing
    that dict and allowing you access through attributes too).

    >>> from wads.util import fc
    >>> list(fc)[20:25]
    ['magenta', 'cyan', 'gray', 'dark_gray', 'dark_red']

    :param string: String with highlight formatting
    :param effect: The effect to use for the highlighting (some unicode like "\033[21m")
    :param beg_mark: String that marks the beginning of the highlight
    :param end_mark: String that marks the end of the highlight
    :param end_effect: The unicode to use to reset the effect
    :return:
    """
    return string.replace(beg_mark, effect).replace(end_mark, end_effect)



import os
import shutil
import json
from collections import ChainMap

# from functools import partial
from typing import List, Optional
from wads import pkg_path_names, root_dir, wads_configs, wads_configs_file
from wads import pkg_join as wads_join
from wads.util import mk_conditional_logger
from wads.pack import write_configs
from wads.licensing import license_body

# from wads.pack_util import write_configs

path_sep = os.path.sep

populate_dflts = wads_configs.get(
    'populate_dflts',
    {
        'description': "There's a bit of an air of mystery around this project...",
        'root_url': None,
        'author': None,
        'license': 'mit',
        'description_file': 'README.md',
        'long_description': 'file:README.md',
        'long_description_content_type': 'text/markdown',
        'keywords': None,
        'install_requires': None,
        'verbose': True,
    },
)


def gen_readme_text(
    name, text="There's a bit of an air of mystery around this project..."
):
    return f'''
# {name}
{text}
'''


# TODO: Add a `defaults_from` in **configs that allows one to have several named defaults in wads_configs_file
def populate_pkg_dir(
    pkg_dir,
    description: str = populate_dflts['description'],
    root_url: Optional[str] = populate_dflts['root_url'],
    author: Optional[str] = populate_dflts['author'],
    license: str = populate_dflts['license'],
    description_file: str = populate_dflts['description_file'],
    keywords: Optional[List] = populate_dflts['keywords'],
    install_requires: Optional[List] = populate_dflts['install_requires'],
    long_description=populate_dflts['long_description'],
    long_description_content_type=populate_dflts[
        'long_description_content_type'
    ],
    include_pip_install_instruction_in_readme=True,
    verbose: bool = populate_dflts['verbose'],
    overwrite: List = (),
    defaults_from: Optional[str] = None,
    skip_docsrc_gen=False,
    **configs,
):
    """Populate project directory root with useful packaging files, if they're missing.

    >>> from wads.populate import populate_pkg_dir
    >>> import os  # doctest: +SKIP
    >>> name = 'wads'  # doctest: +SKIP
    >>> pkg_dir = f'/D/Dropbox/dev/p3/proj/i/{name}'  # doctest: +SKIP
    >>> populate_pkg_dir(pkg_dir,  # doctest: +SKIP
    ...                  description='Tools for packaging',
    ...                  root_url=f'https://github.com/i2mint',
    ...                  author='OtoSense')

    :param pkg_dir:
    :param description:
    :param root_url:
    :param author:
    :param license:
    :param description_file:
    :param keywords:
    :param install_requires:
    :param long_description:
    :param long_description_content_type:
    :param verbose:
    :param default_from: Name of field to look up in wads_configs to get defaults from,
        or 'user_input' to get it from user input.
    :param configs:
    :return:

    """

    args_defaults = dict()
    if defaults_from is not None:
        if defaults_from == 'user_input':  # TODO: Implement!
            args_defaults = dict()  # ... and then fill with user input
            raise NotImplementedError(
                'Not immplemented yet'
            )  # TODO: Implement
        else:
            try:
                wads_configs = json.load(open(wads_configs_file))
                args_defaults = wads_configs[defaults_from]
            except KeyError:
                raise KeyError(
                    f"{wads_configs_file} json didn't have a {defaults_from} field"
                )

    if isinstance(overwrite, str):
        overwrite = {overwrite}
    else:
        overwrite = set(overwrite)

    _clog = mk_conditional_logger(condition=verbose, func=print)
    pkg_dir = os.path.abspath(os.path.expanduser(pkg_dir))
    assert os.path.isdir(pkg_dir), f'{pkg_dir} is not a directory'
    if pkg_dir.endswith(path_sep):
        pkg_dir = pkg_dir[
            :-1
        ]  # remove the slash suffix (or basename will be empty)
    name = os.path.basename(pkg_dir)
    pjoin = lambda *p: os.path.join(pkg_dir, *p)

    if name not in os.listdir(pkg_dir):
        f = pjoin(name)
        _clog(f'... making directory {pkg_dir}')
        os.mkdir(f)
    if '__init__.py' not in os.listdir(pjoin(name)):
        f = pjoin(name, '__init__.py')
        _clog(f'... making an empty {f}')
        with open(f, 'w') as fp:
            fp.write('')

    # Note: Overkill since we just made those things...
    if name not in os.listdir(pkg_dir) or '__init__.py' not in os.listdir(
        pjoin(name)
    ):
        raise RuntimeError(
            "You should have a {name}/{name}/__init__.py structure. You don't."
        )

    if os.path.isfile(pjoin('setup.cfg')):
        with open(pjoin('setup.cfg'), 'r'):
            pass

    kwargs = dict(
        description=description,
        root_url=root_url,
        author=author,
        license=license,
        description_file=description_file,
        long_description=long_description,
        long_description_content_type=long_description_content_type,
        keywords=keywords,
        install_requires=install_requires,
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    # configs = dict(name=name, **configs, **kwargs, **args_defaults)
    # configs = dict(name=name, **args_defaults, **configs, **kwargs)
    configs = dict(ChainMap(dict(name=name), kwargs, configs, args_defaults))

    kwargs['description-file'] = kwargs.pop('description_file', '')

    assert (
        configs.get('name', name) == name
    ), f"There's a name conflict. pkg_dir tells me the name is {name}, but configs tell me its {configs.get('name')}"
    configs['display_name'] = configs.get('display_name', configs['name'])

    def copy_from_resource(resource_name):
        _clog(f'... copying {resource_name} from {root_dir} to {pkg_dir}')
        shutil.copy(wads_join(resource_name), pjoin(resource_name))

    def should_update(resource_name):
        return (resource_name in overwrite) or (
            not os.path.isfile(pjoin(resource_name))
        )

    for resource_name in pkg_path_names:
        if should_update(resource_name):
            copy_from_resource(resource_name)

    def save_txt_to_pkg(resource_name, content):
        target_path = pjoin(resource_name)
        assert not os.path.isfile(target_path), f'{target_path} exists already'
        _clog(f'... making a {resource_name}')
        with open(pjoin(resource_name), 'wt') as fp:
            fp.write(content)

    if should_update('setup.cfg'):
        _clog("... making a 'setup.cfg'")
        if 'pkg-dir' in configs:
            del configs['pkg-dir']
        write_configs(pjoin(''), configs)

    if should_update('LICENSE'):
        _license_body = license_body(configs['license'])
        save_txt_to_pkg('LICENSE', _license_body)

    if should_update('README.md'):
        readme_text = gen_readme_text(name, configs.get('description'))
        if include_pip_install_instruction_in_readme:
            readme_text += f'\n\nTo install:\t```pip install {name}```\n'
        save_txt_to_pkg('README.md', readme_text)

    if not skip_docsrc_gen:
        # TODO: Figure out epythet and wads relationship -- right now, there's a reflexive dependency
        from epythet.setup_docsrc import make_docsrc

        make_docsrc(pkg_dir, verbose)

    return name


def update_pack_and_setup_py(
    target_pkg_dir, copy_files=('setup.py', 'wads/data/MANIFEST.in')
):
    """Just copy over setup.py and pack.py (moving the original to be prefixed by '_'"""
    copy_files = set(copy_files)
    if target_pkg_dir.endswith(path_sep):
        target_pkg_dir = target_pkg_dir[
            :-1
        ]  # remove the slash suffix (or basename will be empty)
    name = os.path.basename(target_pkg_dir)
    contents = os.listdir(target_pkg_dir)
    assert {'setup.py', name}.issubset(
        contents
    ), f"{target_pkg_dir} needs to have all three: {', '.join({'setup.py', name})}"

    pjoin = lambda *p: os.path.join(target_pkg_dir, *p)

    for resource_name in copy_files:
        print(
            f'... copying {resource_name} from {wads_join("")} to {target_pkg_dir}'
        )
        shutil.move(src=pjoin(resource_name), dst=pjoin('_' + resource_name))
        shutil.copy(src=wads_join(resource_name), dst=pjoin(resource_name))


def main():
    import argh  # TODO: replace by argparse, or require argh in wads?

    argh.dispatch_command(populate_pkg_dir)


if __name__ == '__main__':
    main()
